# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
LFER.

Configuration keys:

[lfer]

# glob that catches all LFER files
glob = /path/to/lfer/SA-MARCcomb-lfer-*.tar.gz

# regular expression to get the date out of the path
pattern = ".*SA-MARCcomb-lfer-(?P<year>\d\d)(?P<month>\d\d)(?P<day>\d\d).tar.gz"

"""

from gluish.format import TSV
from gluish.intervals import hourly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.common import Directory
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import iterfiles
import datetime
import json
import luigi
import os
import re

config = Config.instance()

class LFERTask(DefaultTask):
    TAG = '008'

    KIND_FILENAME_MAP = {
        'tit': 'lfer-tit.mrc',
        'lok': 'lfer-lok.mrc',
        'aut': 'lfer-aut.mrc',
    }

    def closest(self):
        if not hasattr(self, 'date'):
            raise RuntimeError('task has not date attribute: %s' % self)
        task = LFERLatestDate(date=self.date)
        luigi.build([task], local_scheduler=True)
        s = task.output().open().read().strip()
        result = datetime.date(*(int(v) for v in s.split('-')))
        return result

class LFERSync(LFERTask):
    """ Copy all LFER files over. """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return Directory(path=os.path.dirname(self.output().path))

    @timed
    def run(self):
        target = os.path.dirname(self.output().path)
        pattern = os.path.join(config.get('lfer', 'glob'))
        shellout("rsync -avz {src} {target}", src=pattern, target=target)
        with self.output().open('w') as output:
            for path in sorted(iterfiles(target)):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class LFERDatesAndPaths(LFERTask):
    """ Just emit a two column TSV with (date, path). """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return LFERSync(indicator=self.indicator)

    @timed
    def run(self):
        pattern = re.compile(config.get('lfer', 'pattern'))
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for row in handle.iter_tsv(cols=('path',)):
                    match = pattern.match(row.path)
                    if match:
                        gd = match.groupdict()
                        year, month, day = map(int, ('20%s' % gd['year'], gd['month'], gd['day']))
                        date = datetime.date(year, month, day)
                        output.write_tsv(date, row.path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class LFERLatestDateAndPath(LFERTask):
    indicator = luigi.Parameter(default=hourly(fmt='%s'))
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return LFERDatesAndPaths(indicator=self.indicator)

    @timed
    def run(self):
        pathmap = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('date', 'path')):
                date = datetime.datetime.strptime(row.date, '%Y-%m-%d').date()
                pathmap[date] = row.path

        def closest_keyfun(date):
            if date > self.date:
                return datetime.timedelta(999999999)
            return abs(date - self.date)

        closest = min(pathmap.keys(), key=closest_keyfun)
        if closest > self.date:
            raise RuntimeError('No shipment before: %s' % min(pathmap.keys()))

        with self.output().open('w') as output:
            output.write_tsv(closest, pathmap.get(closest))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class LFERLatestDate(LFERTask):
    """ Only output the latest date. """
    date = luigi.DateParameter(default=datetime.date.today())
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return LFERLatestDateAndPath(date=self.date, indicator=self.indicator)

    @timed
    def run(self):
        output = shellout("awk '{{print $1}}' {input} > {output}",
                          input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class LFERMarc(LFERTask):
    """ Import a single date or fail if there was no shipment on that date. """
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='tit')

    def requires(self):
        return LFERLatestDateAndPath(date=self.date)

    @timed
    def run(self):
        if self.kind not in self.KIND_FILENAME_MAP:
            raise ValueError('allowed: tit, lok, aut, but got: %s' % self.kind)

        # the filename inside the archive
        filename = self.KIND_FILENAME_MAP.get(self.kind)

        # the latest shipment relative to self.date
        with self.input().open() as handle:
            date, path = handle.iter_tsv(cols=('date', 'path')).next()
        self.logger.debug("The closest shipment occured at {0}".format(date))

        # stream file out
        output = shellout("tar -zOxf {archive} {fn} > {output}",
                          archive=path, fn=filename)
        # move
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class LFERJson(LFERTask):
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='tit')

    def requires(self):
        return LFERMarc(date=self.date, kind=self.kind)

    @timed
    def run(self):
        tmp = shellout("marctojson -m date={date} {input} > {output}",
                       input=self.input().fn, date=self.closest())
        luigi.File(tmp).move(self.output().fn)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class LFERIndex(LFERTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'lfer'
    doc_type = 'title'
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                         '_id': {'path': 'content.001'},
                         '_all': {'enabled': True,
                                  'term_vector': 'with_positions_offsets',
                                  'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return LFERJson(date=self.date, kind='tit')



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from siskin.benchmark import timed
from gluish.common import Executable
from luigi.contrib.esindex import CopyToIndex
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.sources.dbpedia import DBPAbbreviatedNTriples, DBPVirtuoso
from siskin.sources.gnd import GNDAbbreviatedNTriples, GNDVirtuoso
from siskin.task import DefaultTask
import datetime
import luigi
import shutil
import tempfile

class GraphTask(DefaultTask):
    TAG = 'graph'

    def closest(self):
        return monthly(date=self.date)

class GraphCombinedNTriples(GraphTask):
    """ Combine German and English DBpedia and GND. """
    date = ClosestDateParameter(default=datetime.date.today())
    version = luigi.Parameter(default="3.9")

    def requires(self):
        return {'gnd': GNDAbbreviatedNTriples(date=self.date),
                'de': DBPAbbreviatedNTriples(version=self.version, language='de'),
                'en': DBPAbbreviatedNTriples(version=self.version, language='en')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for _, target in self.input().iteritems():
            shellout("cat {input} >> {output}", input=target.path, output=stopover)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='nt'))

class GraphCombinedJson(GraphTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'ntriples': GraphCombinedNTriples(date=self.date),
                'ntto': Executable(name='ntto')}

    @timed
    def run(self):
        output = shellout("ntto -i -j {input} > {output}", input=self.input().get('ntriples').path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class GraphIndex(GraphTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'graph'
    doc_type = 'triple'
    purge_existing_index = True
    mapping = {'triple': {'date_detection': False,
                          'properties': {
                            's': {'type': 'string', 'index': 'not_analyzed'},
                            'p': {'type': 'string', 'index': 'not_analyzed'},
                            'o': {'type': 'string', 'index': 'not_analyzed'}}}}

    def requires(self):
        return GraphCombinedJson(date=self.date)

class GraphCayleyLevelDB(GraphTask):
    """ Create a Cayley LevelDB database from GND and dbpedia data. """
    date = ClosestDateParameter(default=datetime.date.today())
    version = luigi.Parameter(default="3.9")
    language = luigi.Parameter(default="de")

    gomaxprocs = luigi.IntParameter(default=8, significant=False)

    def requires(self):
        return {'ntriples': GraphCombinedNTriples(date=self.date),
                'cayley': Executable(name='cayley', message='http://git.io/KH-wFA')}

    @timed
    def run(self):
        dbpath = tempfile.mkdtemp(prefix='siskin-')
        shellout("cayley init -alsologtostderr -config {config} -dbpath={dbpath}",
             config=self.assets('cayley.leveldb.conf'), dbpath=dbpath)
        shellout("GOMAXPROCS={gomaxprocs} cayley load -config {config} -alsologtostderr -dbpath={dbpath} --triples {input}",
             gomaxprocs=self.gomaxprocs, config=self.assets('cayley.leveldb.conf'), dbpath=dbpath, input=self.input().get('ntriples').path)
        shutil.move(dbpath, self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='leveldb'))

class GraphCayleyMongoDB(GraphTask):
    """ Create a Cayley MongoDB database from GND and dbpedia data. """
    date = ClosestDateParameter(default=datetime.date.today())
    version = luigi.Parameter(default="3.9")
    language = luigi.Parameter(default="de")

    gomaxprocs = luigi.IntParameter(default=8, significant=False)

    def requires(self):
        return {'ntriples': GraphCombinedNTriples(date=self.date),
                'cayley': Executable(name='cayley', message='http://git.io/KH-wFA')}

    @timed
    def run(self):
        shellout("cayley init -alsologtostderr -config {config}",
                 config=self.assets('cayley.mongodb.conf'))
        shellout("cayley load -config {config} -alsologtostderr --triples {input}",
                 config=self.assets('cayley.mongodb.conf'),
                 input=self.input().get('ntriples').path)
        shellout("touch {output}", output=self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mongo'))

#
# Virtuoso related tasks
#
class GraphVirtuoso(GraphTask, luigi.WrapperTask):
    """
    Load GND and selected DBPedia triples into Virtuoso.
    """
    host = luigi.Parameter(default='localhost', significant=False)
    port = luigi.IntParameter(default=1111, significant=False)
    username = luigi.Parameter(default='dba', significant=False)
    password = luigi.Parameter(default='dba', significant=False)

    gnd = luigi.Parameter(default='http://d-nb.info/gnd/')
    date = ClosestDateParameter(default=datetime.date.today())

    dbpedia = luigi.Parameter(default='http://dbpedia.org/resource/')
    version = luigi.Parameter(default="3.9")
    base = luigi.Parameter(default='http://downloads.dbpedia.org',
                           description='base url, for 2014 version, try http://data.dws.informatik.uni-mannheim.de/dbpedia',
                           significant=False)

    def requires(self):
        return {'dbpedia': DBPVirtuoso(host=self.host, port=self.port, username=self.username, password=self.password,
                                       graph=self.dbpedia, version=self.version, base=self.base),
                'gnd': GNDVirtuoso(host=self.host, port=self.port, username=self.username, password=self.password,
                                   graph=self.gnd, date=self.date)}

    def output(self):
        return self.input()



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from siskin.benchmark import timed
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import luigi
import urllib

class ZDBTask(DefaultTask):
    TAG = 'zdb'

    def closest(self):
        return monthly(date=self.date)

class ZDBDump(ZDBTask):
    """ Download the ZDB snapshot. """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='ttl', description='alternatively: rdf')

    @timed
    def run(self):
        host = "datendienst.dnb.de"
        path = "/cgi-bin/mabit.pl"
        params = {
            "userID": "opendata",
            "pass": "opendata",
            "cmd": "fetch",
            "mabheft": "ZDBTitel.%s.gz" % self.format,
        }

        url = 'http://{host}{path}?{params}'.format(host=host,
                                                    path=path,
                                                    params=urllib.urlencode(params))
        output = shellout("""wget --retry-connrefused "{url}" -O {output}""", url=url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='%s.gz' % self.format))

class ZDBExtract(ZDBTask):
    """ Extract the archive. """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='ttl', description='alternatively: rdf')

    def requires(self):
        return ZDBDump(date=self.date, format=self.format)

    @timed
    def run(self):
        output = shellout("gunzip -c {input} > {output}", input=self.input().fn)
        luigi.File(output).move(self.output().fn)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext=self.format))

class ZDBNTriples(ZDBTask):
    """ Get a Ntriples representation of ZDB. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return ZDBExtract(date=self.date)

    @timed
    def run(self):
        output = shellout("serdi -b -i turtle -o ntriples {input} > {output}",
                          input=self.input().path)
        # output = shellout("sort {input} > {output}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='nt'))



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Possible mitigation of https://github.com/CrossRef/rest-api-doc/issues/67.
"""

from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.workflows.ai import AIIntermediateSchema
import datetime
import json
import luigi

class CRFixTask(DefaultTask):
    TAG = 'crfix'

    def closest(self):
        return monthly(date=self.date)

class CRFixTitles(CRFixTask):
    """
    Extract all article titles from intermediate schema.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AIIntermediateSchema(date=self.date)

    def run(self):
        output = shellout("""jq -r '.["rft.atitle"]' <(unpigz -c {input}) > {output}""", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class CRFixAuthors(CRFixTask):
    """
    Extract author names.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AIIntermediateSchema(date=self.date)

    def run(self):
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    c = json.loads(line)
                    authors = c.get("authors")
                    if authors:
                        for author in authors:
                            if author.get("rft.au"):
                                output.write_tsv(author["rft.au"].encode('utf-8'))
                                continue
                            if author.get("rft.aulast") and author.get("rft.aufirst"):
                                name = author["rft.aufirst"].encode('utf-8') + " " + author["rft.aulast"].encode('utf-8')
                                output.write_tsv(name)
                                continue
                            if author.get("rft.aulast"):
                                output.write_tsv(author["rft.aulast"].encode('utf-8'))
                                continue
                            if author.get("rft.aufirst"):
                                output.write_tsv(author["rft.aufirst"].encode('utf-8'))
                                continue

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Nationallizenzen.

Configuration keys:

[core]

swb-mirror = /path/to/swb-mirror

google-username = user.name
google-password = secret
google-docs-key = crypticstring

"""

from gluish.format import TSV
from gluish.intervals import hourly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.common import Directory
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import iterfiles
import datetime
import gspread
import json
import logging
import luigi
import os
import re
import string
import sys
import tempfile

config = Config.instance()

class NLTask(DefaultTask):
    TAG = '017'

    def closest(self):
        if not hasattr(self, 'date'):
            raise RuntimeError('task has no date: %s' % self)
        task = NLLatestPackageDate(date=self.date)
        luigi.build([task], local_scheduler=True)
        s = task.output().open().read().strip()
        result = datetime.date(*(int(v) for v in s.split('-')))
        return result

class NLSync(NLTask):
    """ Sync the complete 'nationallizenzen' folder as is. """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return Directory(path=os.path.dirname(self.output().path))

    def run(self):
        source = os.path.join(config.get('core', 'swb-mirror'), 'nationallizenzen')
        target = os.path.dirname(self.output().path)
        shellout("rsync -avz {source} {target}", source=source, target=target)
        with self.output().open('w') as output:
            for path in iterfiles(target):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class NLInventory(NLTask):
    """ All (date, path) for a section, type, format and tag. """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))
    section = luigi.Parameter(default='Monografien',
                              description='Monographien, ZDB, ZDB_in_Ein...')
    type = luigi.Parameter(default='SA', description='SA or SZ')
    format = luigi.Parameter(default='MARC', description='MARC, PICA, OLIX, '
                             'MAB2, MABPPN, MABPPNOS, MABPPNOScomb, MARCcomb')
    tag = luigi.Parameter(default='NL', description='NL or ILN')

    def requires(self):
        return NLSync(indicator=self.indicator)

    def run(self):
        template = (r'.*nationallizenzen/{0.section}/{0.type}-{0.format}-{0.tag}-'
                    r'(\d{{2}})(\d{{2}})(\d{{2}}).tar.gz')
        pattern = template.format(self)
        events = []

        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                match = re.search(pattern, row.path)
                if match:
                    year, month, day = map(int, match.groups())
                    year += 2000
                    date = datetime.date(year, month, day)
                    events.append((date, row.path))

        with self.output().open('w') as output:
            for date, path in sorted(events):
                output.write_tsv(date, path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class NLLatestDateAndPath(NLTask):
    """ The latest (date, path) for a section, type, format and tag. """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))
    date = luigi.DateParameter(default=datetime.date.today())
    section = luigi.Parameter(default='Monografien',
                              description='Monographien, ZDB, ZDB_in_Ein...')
    type = luigi.Parameter(default='SA', description='SA or SZ')
    format = luigi.Parameter(default='MARC', description='MARC, PICA, OLIX, '
                             'MAB2, MABPPN, MABPPNOS, MABPPNOScomb, MARCcomb')
    tag = luigi.Parameter(default='NL', description='NL or ILN')

    def requires(self):
        return NLInventory(indicator=self.indicator, section=self.section,
                           type=self.type, format=self.format, tag=self.tag)

    def run(self):
        pathmap = {}
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('date', 'path')):
                date = datetime.datetime.strptime(row.date, '%Y-%m-%d').date()
                pathmap[date] = row.path

        def closest_keyfun(date):
            if date > self.date:
                return datetime.timedelta(999999999)
            return abs(date - self.date)

        closest = min(pathmap.keys(), key=closest_keyfun)
        if closest > self.date:
            raise RuntimeError('No shipment before: %s for %s' % (
                               min(pathmap.keys()), self))

        with self.output().open('w') as output:
            output.write_tsv(closest, pathmap.get(closest))

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class NLPackageDescriptor(NLTask):
    """ Download package descriptor from google docs. """
    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def run(self):
        gc = gspread.login(config.get('core', 'google-username'),
                           config.get('core', 'google-password'))
        doc = gc.open_by_key(config.get('core', 'google-docs-key'))
        sheet = doc.get_worksheet(4)
        with self.output().open('w') as output:
            list_of_lists = iter(sheet.get_all_values())
        list_of_lists.next()
        with self.output().open('w') as output:
            for row in list_of_lists:
                section, type, format, tag = map(string.strip, row)
                output.write_tsv(section, type, format, tag)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class NLLatestPackageDate(NLTask):
    """ Return the latest date of some combined package. """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        prerequisite = NLPackageDescriptor()
        luigi.build([prerequisite], local_scheduler=True)
        with prerequisite.output().open() as handle:
            for row in handle.iter_tsv(cols=('section', 'type', 'fmt', 'tag')):
                yield NLLatestDateAndPath(date=self.date, section=row.section,
                                          type=row.type, format=row.fmt,
                                          tag=row.tag)

    def run(self):
        dates = set()
        for target in self.input():
            date, _ = target.open().next().strip().split('\t', 1)
            dates.add(date)
        with self.output().open('w') as output:
            output.write_tsv(max(dates))

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class NLPackage(NLTask):
    """ A custom combined NL shipment. """
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='tit', description='tit, lok or aut')

    def requires(self):
        prerequisite = NLPackageDescriptor()
        luigi.build([prerequisite], local_scheduler=True)
        with prerequisite.output().open() as handle:
            for row in handle.iter_tsv(cols=('section', 'type', 'fmt', 'tag')):
                yield NLLatestDateAndPath(date=self.date, section=row.section,
                                          type=row.type, format=row.fmt,
                                          tag=row.tag)

    def run(self):
        _, combined = tempfile.mkstemp(prefix='tasktree-')
        for target in self.input():
            with target.open() as handle:
                _, path = handle.iter_tsv(cols=('date', 'path')).next()
                if sys.platform.startswith("linux"):
                    shellout("tar -O -zf {input} -x --wildcards --no-anchored '*-{kind}.mrc' >> {output}",
                         kind=self.kind, input=path, output=combined)
                if sys.platform == "darwin":
                    shellout("tar -O -zf {input} -x --include='*-{kind}.mrc' >> {output}",
                         kind=self.kind, input=path, output=combined)
        luigi.File(combined).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class NLJson(NLTask):
    date = ClosestDateParameter(default=datetime.date.today())
    kind = luigi.Parameter(default='tit', description='tit, lok or aut')

    def requires(self):
        return NLPackage(date=self.date, kind=self.kind)

    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class NLIndex(NLTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'nl'
    doc_type = 'title'
    purge_existing_index = True

    settings = {
        "settings": {
            "number_of_shards": 5,
            "analysis": {
                "filter": {
                    "autocomplete_filter": {
                        # normal ngram vs edge?
                        "type":     "edge_ngram",
                        "min_gram": 1,
                        "max_gram": 20
                    }
                },
                "analyzer": {
                    "autocomplete": {
                        "type":      "custom",
                        "tokenizer": "standard",
                        "filter": [
                            # add more?
                            "lowercase",
                            "autocomplete_filter"
                        ]
                    }
                }
            }
        }
    }

    mapping = {
        'title': {
            'date_detection': False,
            '_id': {
                'path': 'content.001'
            },
            '_all': {
                'enabled': True,
                'term_vector': 'with_positions_offsets',
                'store': True
            },
            'properties': {
                'content': {
                    'properties': {
                        '245': {
                            'properties': {
                                # multifield?
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        },
                        '100': {
                            'properties': {
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        },
                        '700': {
                            'properties': {
                                'a': {
                                    'type': 'string',
                                    'index_analyzer': 'autocomplete',
                                    'search_analyzer': 'standard'
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return NLJson(date=self.date, kind='tit')



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
GBV related tasks.

Configutation keys:

[gbv]

scp-src = [user@server:[port]]/path/to/file/{tag}/*mrc

Requirements:

    * marcuniq, marctojson
    * elasticsearch
"""

from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import iterfiles
import datetime
import luigi
import tempfile

config = Config.instance()

class GBVTask(DefaultTask):
    """ GBV base task. """
    TAG = '021'

    def closest(self):
        """ Singular event. """
        return datetime.date(2014, 1, 1)

class GBVImport(GBVTask):
    """ Copy and merge files from a single tag on the fly. Tag is the name of
    the subdirectory, like haab, hfm or sbb.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    tag = luigi.Parameter(default='haab')

    @timed
    def run(self):
        stopover = tempfile.mkdtemp(prefix='siskin-')
        origin = config.get('gbv', 'scp-src').format(tag=self.tag)
        shellout("scp {origin} {output}", origin=origin, output=stopover)

        _, combined = tempfile.mkstemp(prefix='siskin-')
        for path in iterfiles(stopover):
            shellout("cat {input} >> {output}", input=path, output=combined)
        luigi.File(combined).move(self.output().fn)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class GBVCombine(GBVTask):
    """ Combine different files from different tags together. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        for tag in ('haab', 'hfm', 'sbb'):
            yield GBVImport(date=self.date, tag=tag)

    @timed
    def run(self):
        _, combined = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}", input=target.path,
                     output=combined)
        output = shellout("marcuniq {input} > {output}", input=combined)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class GBVJson(GBVTask):
    """ Convert to Json. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GBVCombine(date=self.date)

    @timed
    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class GBVIndex(GBVTask, CopyToIndex):
    """ Upload to ES. """
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'gbv'
    doc_type = 'title'
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return GBVJson(date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
HMT related workflows.

Data reconciliation, with target artefacts and precision reports.
"""

from __future__ import print_function
from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.sources.naxos import NaxosJson
from siskin.task import DefaultTask
from siskin.utils import ElasticsearchMixin
import datetime
import itertools
import json
import luigi


class HMTTask(DefaultTask):
    TAG = 'hmt'
    NULL = '<NULL>'

    def closest(self):
        return weekly(date=self.date)

class HMTNaxosPersonsAndDates(HMTTask):
    """
    Dump all people and their bio data from 100 and 700 fields.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    encoding = luigi.Parameter(default='utf-8', significant=False)

    def requires(self):
        return NaxosJson(date=self.date)

    def run(self):
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for _, line in enumerate(handle):
                    c = json.loads(line).get('content')
                    id = c.get('001', self.NULL)
                    for field in itertools.chain(c.get('100', []), c.get('700', [])):
                        output.write_tsv(id, field.get('a').encode(self.encoding), field.get('d', self.NULL))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class HMTNaxosUniqPersonsAndDates(HMTTask):
    """ Return the unique authors. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return HMTNaxosPersonsAndDates(date=self.date)

    def run(self):
        output = shellout("cut -f 2-3 < {input} | LANG=C sort -u > {output}",
                          input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class HMTNaxosCompletePersonsAndDates(HMTTask):
    """
    Only names with dates.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return HMTNaxosUniqPersonsAndDates(date=self.date)

    def run(self):
        output = shellout("""LANG=C grep -v "{null}" {input} > {output}""",
                          null=self.NULL, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class HMTNaxosIncompleteAuthorsAndDates(HMTTask):
    """
    Only names without dates.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return HMTNaxosUniqPersonsAndDates(date=self.date)

    def run(self):
        output = shellout("""LANG=C grep "{null}" {input} > {output}""",
                          null=self.NULL, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
MOR.

Configuration keys:

[mor]

ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-base = .
ftp-pattern = *

"""

from siskin.benchmark import timed
from siskin.common import FTPMirror
from luigi.contrib.esindex import CopyToIndex
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
import datetime
import luigi
import tempfile

config = Config.instance()

class MORTask(DefaultTask):
    TAG = '006'

    def closest(self):
        return datetime.date(2014, 1, 1)

class MORJson(MORTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return FTPMirror(host=config.get('mor', 'ftp-host'),
                         username=config.get('mor', 'ftp-username'),
                         password=config.get('mor', 'ftp-password'),
                         base=config.get('mor', 'ftp-base'),
                         pattern=config.get('mor', 'ftp-pattern'))

    @timed
    def run(self):
        _, combined = tempfile.mkstemp(prefix='tasktree-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if row.path.endswith('.xml'):
                    output = shellout("""yaz-marcdump -i marcxml -o marc
                                      {input} >> {output}""", input=row.path,
                                      output=combined)
        output = shellout("""marcuniq {input} > {output}""", input=combined)
        output = shellout("""marctojson -m date={date} {input} > {output}""",
                          date=self.closest(), input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class MORIndex(MORTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'mor'
    doc_type = 'title'
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return MORJson(date=self.date)



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                    The Finc Authors, http://finc.info
#                    Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#
#
# Microsoft Academic Graph
#
# TOS: http://research.microsoft.com/en-us/um/redmond/projects/mag/license.txt
#
# Links:
#
# * http://research.microsoft.com/en-us/projects/mag/
# * https://academicgraph.blob.core.windows.net/graph/index.html?eulaaccept=on
#
# Number of papers as of November 2015: 120887833, ~37M with DOI
#
# Schema
# Affiliations
#
#     Affiliation ID
#     Affiliation name
#
# Authors
#
#     Author ID
#     Author name
#
# Conference
#
#     Conference ID
#     Short name (abbreviation)
#     Full name
#
# ConferenceInstances
#
#     Conference ID
#     Conference instance ID
#     Short name (abbreviation)
#     Full name
#     Location
#     Official conference URL
#     Conference start date
#     Conference end date
#     Conference abstract registration date
#     Conference submission deadline date
#     Conference notification due date
#     Conference final version due date
#
# FieldsOfStudy
#
#     Field of study ID
#     Field of study name
#
# Journals
#
#     Journal ID
#     Journal name
#
# Papers
#
#     Paper ID
#     Original paper title
#     Normalized paper title
#     Paper publish year
#     Paper publish date
#     Paper Document Object Identifier (DOI)
#     Original venue name
#     Normalized venue name
#     Journal ID mapped to venue name
#     Conference series ID mapped to venue name
#     Paper rank
#
# PaperAuthorAffiliations
#
#     Paper ID
#     Author ID
#     Affiliation ID
#     Original affiliation name
#     Normalized affiliation name
#     Author sequence number
#
# PaperKeywords
#
#     Paper ID
#     Keyword name
#     Field of study ID mapped to keyword
#
# PaperReferences
#
#     Paper ID
#     Paper reference ID
#
# PaperUrls
#
#     Paper ID
#     URL

from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
import datetime
import luigi

config = Config.instance()

class MAGTask(DefaultTask):
    TAG = 'mag'

    def closest(self):
        return datetime.date(2015, 11, 6)

class MAGDates(MAGTask):
    """ Extract dates. """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""
            curl "https://academicgraph.blob.core.windows.net/graph/index.html?eulaaccept=on" |
            grep "<li>" | grep -o "20[0-9][0-9]-[01][1-9]-[012][0-9]" | sort -u > {output} """)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class MAGDump(MAGTask):
    """ Dump file. """

    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""
            curl https://academicgraph.blob.core.windows.net/graph-{date}/MicrosoftAcademicGraph.zip > {output}""",
            date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="zip"))

class MAGFile(MAGTask):
    """
    Listing archive: MicrosoftAcademicGraph.zip

    --
    Path = MicrosoftAcademicGraph.zip
    Type = zip
    64-bit = +
    Physical Size = 29698421155

       Date      Time    Attr         Size   Compressed  Name
    ------------------- ----- ------------ ------------  ------------------------
    2015-11-16 15:45:20 ....A       770185       296305  Affiliations.txt
    2015-11-16 15:48:48 ....A   3037120573   1472531330  Authors.txt
    2015-11-16 15:45:22 ....A     10346759      3502468  ConferenceInstances.txt
    2015-11-16 15:45:20 ....A        79774        26235  Conferences.txt
    2015-11-16 15:45:20 ....A      1508780       713334  FieldsOfStudy.txt
    2015-11-16 15:45:22 ....A      1002313       369790  Journals.txt
    2015-06-11 22:34:30 ....A         9641         3998  license.txt
    2015-11-16 16:03:48 ....A  18017900470   4298547230  PaperAuthorAffiliations.txt
    2015-11-16 15:51:32 ....A   5329643052   1839341847  PaperKeywords.txt
    2015-11-16 16:21:14 ....A  18094921016   7421585112  PaperReferences.txt
    2015-11-16 16:27:50 ....A  27419818455   8967796102  Papers.txt
    2015-11-16 16:22:22 ....A  27259393790   5693705196  PaperUrls.txt
    2015-08-30 17:21:24 ....A         1309          484  readme.txt
    ------------------- ----- ------------ ------------  ------------------------
                               99172516117  29698419431  13 files, 0 folders
    """
    date = ClosestDateParameter(default=datetime.date.today())
    name = luigi.Parameter(default='Papers')

    def requires(self):
        return MAGDump(date=self.date)

    def run(self):
        output = shellout("7z e -so {dump} {name}.txt | pigz -c > {output}", dump=self.input().path, name=self.name)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"))

class MAGPaperDomains(MAGTask):
    """
    Domain distribution of Paper URLs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MAGFile(name='PaperUrls', date=self.date)

    def run(self):
        output = shellout("""
            unpigz -c {file} | LC_ALL=C grep -o "http[s]*://[^/]*" |
            LC_ALL=C sed -e 's@http[s]*://@@g' | TMPDIR={tmpdir} LC_ALL=C sort -S50% |
            uniq -c | sort -nr > {output} """,
            tmpdir=config.get('core', 'tempdir'), file=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"))

class MAGDOIList(MAGTask):
    """
    List of DOI in this dataset.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MAGFile(date=self.date, name='Papers')

    def run(self):
        output = shellout(""" unpigz -c {input} | LC_ALL=C cut -f6 | LC_ALL=C grep -v ^$ | LC_ALL=C sort -S50% | pigz -c > {output} """, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"))

class MAGKeywordDistribution(MAGTask):
    """
    Keyword distribution.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MAGFile(date=self.date, name='PaperKeywords')

    def run(self):
        output = shellout("unpigz -c {input} | cut -f2 | TMPDIR={tmpdir} sort -S50% | uniq -c | sort -nr | pigz -c > {output}",
                          input=self.input().path, tmpdir=config.get('core', 'tempdir'))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Music Treasury Collection.
"""

from gluish.format import TSV
from gluish.intervals import quarterly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from lxml import etree
from siskin.benchmark import timed
from siskin.task import DefaultTask
from siskin.utils import iterfiles, wc
import BeautifulSoup
import cStringIO
import datetime
import luigi
import os
import tempfile

class MTCTask(DefaultTask):
    TAG = '010'

    def closest(self):
        """ Try redownload every three month. """
        return quarterly(date=self.date)

class MTCSitemap(MTCTask):
    """ Download the sitemap. """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default='http://lcweb2.loc.gov/diglib/ihas/consortium2/sitemap.xml', significant=False)

    @timed
    def run(self):
        output = shellout("wget -q --retry-connrefused -O {output} {url}", url=self.url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))

class MTCStylesheet(MTCTask):
    """ Download MODS -> MARC21 XSL stylesheet. TODO: this could be added to
    the assets. """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default='http://www.loc.gov/standards/marcxml/xslt/MODS2MARC21slim.xsl', significant=False)

    @timed
    def run(self):
        output = shellout("""wget -q --retry-connrefused -O {output} {url}""", url=self.url)
        output = shellout(""" sed -e 's@include href="http://www.loc.gov/marcxml/xslt/MARC21slimUtils.xsl"@include href="{local}"@' {input} > {output}""", input=output, local=self.assets("MARC21slimUtils.xsl"))
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xsl'))

class MTCUrls(MTCTask):
    """ Extract URLs from MTC sitemap. The sitemap just contains links
    to HTML files, but the path to the metadata is almost the same.

    #TODO: maybe via XSLT, too?
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MTCSitemap(date=self.date)

    @timed
    def run(self):
        with self.input().open('r') as handle:
            soup = BeautifulSoup.BeautifulStoneSoup(handle.read())
            locs = soup.findAll('loc')
            with self.output().open('w') as output:
                for loc in locs:
                    url = loc.text.replace('default.html', 'mods.xml')
                    if url.endswith('.xml'):
                        output.write_tsv(url)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class MTCModsImport(MTCTask):
    """ Download all files sequentially. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MTCUrls(date=self.date)

    @timed
    def run(self):
        # create target subdirectory
        target = os.path.join(os.path.dirname(self.output().path), str(self.closest()))
        if not os.path.exists(target):
            os.makedirs(target)
        size = wc(self.input().path)

        with self.input().open() as handle:
            for i, row in enumerate(handle.iter_tsv(cols=('url',)), start=1):
                name = os.path.join(target, row.url.split('/')[-2])
                destination = "{name}.xml".format(name=name)
                if not os.path.exists(destination):
                    output = shellout("""wget -q --retry-connrefused
                                      {url} -O {output}""", url=row.url)
                    luigi.File(output).move(destination)
                self.logger.debug("{0}/{1} {2}".format(i, size, row.url))

        # write "receipt"
        with self.output().open('w') as output:
            for path in iterfiles(target):
                if path.endswith('.xml'):
                    output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class MTCMarcXML(MTCTask):
    """ Convert the .mods XML files to Marc XML files via `xsltproc`.
    Again, the completion is indicated by a file containing all pathes
    to the converted records.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'stylesheet': MTCStylesheet(date=self.date),
                'filelist': MTCModsImport(date=self.date)}

    @timed
    def run(self):
        target = os.path.join(os.path.dirname(self.output().path), str(self.date))
        if not os.path.exists(target):
            os.makedirs(target)

        _, errorlog = tempfile.mkstemp(prefix='siskin-')
        stylesheet = self.input().get('stylesheet').path
        size = wc(self.input().get('filelist').path)

        with self.input().get('filelist').open() as handle:
            for i, row in enumerate(handle.iter_tsv(cols=('path',)), start=1):
                basename = os.path.basename(row.path)
                name = basename.replace(".xml", ".marcxml")
                destination = os.path.join(target, name)
                if not os.path.exists(destination):
                    try:
                        output = shellout("xsltproc {xsl} {input} > {output}",
                                          input=row.path, xsl=stylesheet)
                        luigi.File(output).move(destination)
                    except RuntimeError as err:
                        self.logger.error("{0}: {1}".format(row.path, err))
                        with open(errorlog, 'a') as log:
                            log.write('%s\t%s\n' % (row.path, err))
                self.logger.debug("{0}/{1} {2}".format(i, size, row.path))

        # write receipt
        with self.output().open('w') as output:
            for path in iterfiles(target):
                output.write_tsv(path)

        # this is just a temporary artefact for now
        self.logger.debug("Conversion errors logged at: {0}".format(errorlog))

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class MTCCombine(MTCTask):
    """ Convert Marc XML to Marc8, via `yaz-marcdump`.
    yaz-marcdump version 4.X returned 0 to the shell on yaz_marc_read_xml failed,
    whereas version 5.X exits with a non-zero code (5).
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MTCMarcXML(date=self.date)

    @timed
    def run(self):
        target = os.path.join(os.path.dirname(self.output().path), str(self.date))
        if not os.path.exists(target):
            os.makedirs(target)

        size = wc(self.input().path)
        _, combined = tempfile.mkstemp(prefix='siskin-')

        with self.input().open() as handle:
            for i, row in enumerate(handle.iter_tsv(cols=('path',)), start=1):

                # Cleanup wrongly nested data fields, see:
                # https://gist.github.com/miku/ea779a221d00b5524fcd
                # in 2014-05, this corrects 673 errors, while 31 are not yet
                # recoverable!
                with open(row.path) as handle:
                    f = cStringIO.StringIO(handle.read())
                    doc = etree.parse(f)

                result = doc.xpath('/marc:record/marc:datafield/marc:datafield',
                          namespaces={'marc': 'http://www.loc.gov/MARC21/slim'})
                if len(result) > 0:
                    self.logger.debug("Fixing broken MARCXML in: {0}".format(row.path))
                for misplaced in result:
                    parent = misplaced.getparent()
                    record = misplaced.getparent().getparent()
                    parent.remove(misplaced)
                    record.append(misplaced)

                _, cleaned = tempfile.mkstemp(prefix='siskin-')
                with open(cleaned, 'w') as output:
                    output.write(etree.tostring(doc, pretty_print=True))

                # actually do the conversion ...
                basename = os.path.basename(row.path)
                name = basename.replace(".marcxml", ".mrc")
                destination = os.path.join(target, name)
                if not os.path.exists(destination):
                    # exit(5) for serious decoding errors
                    # see: http://www.indexdata.com/yaz/doc/NEWS
                    shellout("""yaz-marcdump -i marcxml -o marc
                             {input} >> {output}""", input=cleaned,
                             output=combined, ignoremap={5: 'FIXME'})
                self.logger.debug("{0}/{1} {2}".format(i, size, row.path))

        luigi.File(combined).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class MTCJson(MTCTask):
    """ Convert XML files to a single JSON file."""
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MTCCombine(date=self.date)

    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                       date=self.closest(), input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class MTCIndex(MTCTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'mtc'
    doc_type = 'title'
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return MTCJson(date=self.date)



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#
#
# Aggregated Index workflows
# ==========================
#
# AMSL (api) + AIIntermediateSchema (sources) = AILicensing (isil) -> AIExport (for solr)
#

from BeautifulSoup import BeautifulSoup
from gluish.common import Executable
from gluish.format import TSV, Gzip
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.configuration import Config
from siskin.sources.amsl import AMSLHoldingsFile, AMSLCollections, AMSLHoldingsISILList, AMSLCollectionsISIL
from siskin.sources.crossref import CrossrefIntermediateSchema, CrossrefUniqISSNList
from siskin.sources.degruyter import DegruyterIntermediateSchema, DegruyterISSNList
from siskin.sources.doaj import DOAJIntermediateSchema, DOAJISSNList
from siskin.sources.gbi import GBIIntermediateSchemaByKind
from siskin.sources.gbi import GBIISSNList
from siskin.sources.holdings import HoldingsFile
from siskin.sources.jstor import JstorIntermediateSchema, JstorISSNList
from siskin.task import DefaultTask
from siskin.utils import URLCache
import collections
import datetime
import itertools
import json
import luigi
import os
import re
import requests
import string
import tempfile

config = Config.instance()

class AITask(DefaultTask):
    TAG = 'ai'

    def closest(self):
        return weekly(self.date)

class DownloadFile(AITask):
    """ Download a file. TODO(miku): move this out here. """

    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter()

    @timed
    def run(self):
        output = shellout("""curl -L --fail "{url}" > {output}""", url=self.url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='file', digest=True))

class DownloadAndUnzipFile(AITask):
    """ Download a file and unzip it. TODO(miku): move this out here. """
    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter()

    @timed
    def run(self):
        output = shellout("""curl -L --fail "{url}" > {output}""", url=self.url)
        output = shellout("""unzip -qq -p {input} > {output}""", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='file', digest=True))

class AIISSNStats(AITask):
    """ Match ISSN lists. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date)
        }

    @timed
    def run(self):
        def loadset(target):
            s = set()
            with target.open() as handle:
                for row in handle.iter_tsv(cols=('issn',)):
                    s.add(row.issn)
            return s

        with self.output().open('w') as output:
            for k1, k2 in itertools.combinations(self.input().keys(), 2):
                s1 = loadset(self.input().get(k1))
                s2 = loadset(self.input().get(k2))
                output.write_tsv(k1, k2, len(s1), len(s2), len(s1.intersection(s2)))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIISSNOverlaps(AITask):
    """ Match ISSN lists. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            # 'gbi': GBIISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date)
        }

    @timed
    def run(self):
        def loadset(target):
            s = set()
            with target.open() as handle:
                for row in handle.iter_tsv(cols=('issn',)):
                    s.add(row.issn)
            return s

        with self.output().open('w') as output:
            for k1, k2 in itertools.combinations(self.input().keys(), 2):
                s1 = loadset(self.input().get(k1))
                s2 = loadset(self.input().get(k2))
                for issn in sorted(s1.intersection(s2)):
                    output.write_tsv(k1, k2, issn)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIIntermediateSchema(AITask):
    """ Create an intermediate schema record from all AI sources. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        """
        TODO(miku): Dynamic dependencies based on holding file.
        """
        return [Executable(name='pigz', message='http://zlib.net/pigz/'),
                GBIIntermediateSchemaByKind(kind='references', date=self.date),
                GBIIntermediateSchemaByKind(kind='fulltext', date=self.date),
                CrossrefIntermediateSchema(date=self.date),
                DegruyterIntermediateSchema(date=self.date),
                DOAJIntermediateSchema(date=self.date),
                JstorIntermediateSchema(date=self.date)]

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input()[1:]:
            shellout("cat {input} >> {output}", input=target.path, output=stopover)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)

class AIFilterConfigISIL(AITask):
    """
    Next iteration of filtering, based on various AMSL responses.
    Creates the filters for a single ISIL from AMSL.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(description='ISIL')

    def requires(self):
        return AMSLCollectionsISIL(isil=self.isil)

    @timed
    def run(self):
        filters = []
        with self.input().open() as handle:
            c = json.load(handle)

        for source, colls in c.iteritems():
            filters.append({'and': [{'source': source}, {'collections': colls}]})

        with self.output().open('w') as output:
            output.write(json.dumps({self.isil: {'or': filters}}))

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class AIFilterConfigNext(AITask):
    """
    Next iteration of filtering, based on various AMSL responses.
    ISILs are hard-coded, but should come from the API as well.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'DE-15': AIFilterConfigISIL(isil='DE-15'),
            'DE-14': AIFilterConfigISIL(isil='DE-14'),
        }

    @timed
    def run(self):
        config = {}
        for isil, target in self.input().iteritems():
            with target.open() as handle:
                c = json.load(handle)
                config[isil] = c[isil]

        with self.output().open('w') as output:
            output.write(json.dumps(config))

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class AIFilterConfig(AITask):
    """
    Create a filter configuration from AMSL.

    The filterconfig dictionary should be built from AMSL data only: holdings
    and collections, and information about which source or collection uses
    which method of labeling.

    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        """
        Intermediate schema files and holdings.
        """
        return {
            # holding files
            'DE-105': AMSLHoldingsFile(isil='DE-105'),
            'DE-14': AMSLHoldingsFile(isil='DE-14'),
            'DE-15': AMSLHoldingsFile(isil='DE-15'),
            'DE-1972': AMSLHoldingsFile(isil='DE-1972'),
            'DE-8': AMSLHoldingsFile(isil='DE-8'),
            'DE-Bn3': AMSLHoldingsFile(isil='DE-Bn3'),
            'DE-Brt1': AMSLHoldingsFile(isil='DE-Brt1'),
            'DE-Ch1': AMSLHoldingsFile(isil='DE-Ch1'),
            'DE-D117': AMSLHoldingsFile(isil='DE-D117'),
            'DE-D161': AMSLHoldingsFile(isil='DE-D161'),
            'DE-Gla1': AMSLHoldingsFile(isil='DE-Gla1'),
            'DE-J59': AMSLHoldingsFile(isil='DE-J59'),
            'DE-Ki95': AMSLHoldingsFile(isil='DE-Ki95'),
            'DE-Rs1': AMSLHoldingsFile(isil='DE-Rs1'),
            'DE-Zi4': AMSLHoldingsFile(isil='DE-Zi4'),
            'DE-Kn38': AMSLHoldingsFile(isil='DE-Kn38'),

            # issn list
            'DE-15-FID': DownloadFile(date=self.date, url='https://goo.gl/tm6U9D'),
        }

    @timed
    def run(self):
        """
        Build a filter from a template. Fill in the missing files.
        """
        with open(self.assets('filterconf.template.json')) as handle:
            template = json.load(handle)

        config = {}

        for isil, filtertree in template.iteritems():
            template = json.dumps(filtertree)
            if "{{ file }}" in template and isil not in self.input():
                raise RuntimeError('isil %s seems to expect input file, but none given as dependency' % isil)
            tree = template.replace("{{ file }}", self.input().get(isil).path)
            config[isil] = json.loads(tree)

        with self.output().open('w') as output:
            output.write(json.dumps(config))

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='json'))

class AILicensing(AITask):
    """
    Take intermediate schema and a config and attach ISILs accordingly.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'is': AIIntermediateSchema(date=self.date),
            'config': AIFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('is').path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))

class AIDuplicates(AITask):
    """
    First attempt at a list of id, DOI, ISIL list for simple duplicate detection.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AILicensing(date=self.date)

    def run(self):
        output = shellout("""jq -r '[.["finc.record_id"]?, .doi?, .["rft.atitle"]?, .["x.labels"][]?] | @csv' <(unpigz -c {input}) > {output} """,
                          input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIDuplicatesSortbyDOI(AITask):
    """
    Sort list by DOI.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AIDuplicates(date=self.date)

    def run(self):
        output = shellout("""LC_ALL=C sort -t , -k2,2 -S35% {input} > {output} """, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIExport(AITask):
    """
    Next iteration of AI export.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return AILicensing(date=self.date)

    def run(self):
        output = shellout("span-export <(unpigz -c {input}) | pigz -c > {output}", input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))

class AIUpdate(AITask, luigi.WrapperTask):
    """
    Just a wrapper task for updates, refs #5702.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [AIExport(date=self.date), AIIntermediateSchema(date=self.date)]

    def output(self):
        return self.input()

class AICoverageISSN(AITask):
    """
    For all AI sources and all holding files, find the number
    (perc) of ISSNs in the holding file, that we have some reach to.

    Result looks something like this, indicating availablity and sources:

    2091-2145   crossref|doaj
    2091-2234   crossref
    2091-2560   crossref
    2091-2609   crossref
    2091-2730   NOT_FOUND

    """
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def requires(self):
        """
        TODO: Add GBI ISSN list.
        """
        return {
            'crossref': CrossrefUniqISSNList(date=self.date),
            'jstor': JstorISSNList(date=self.date),
            'degruyter': DegruyterISSNList(date=self.date),
            'doaj': DOAJISSNList(date=self.date),
            'gbi': GBIISSNList(date=self.date),
            'file': AMSLHoldingsFile(isil=self.isil),
        }

    def run(self):
        issns = collections.defaultdict(set)

        with self.input().get('file').open() as handle:
            for row in handle:
                fields = row.strip().split('\t')
                if len(fields) < 3:
                    continue

                if re.search(r'[0-9]{4}-[0-9]{3}[0-9X]', fields[1]):
                    issns['file'].add(fields[1])

                if re.search(r'[0-9]{4}-[0-9]{3}[0-9X]', fields[2]):
                    issns['file'].add(fields[2])

        sources = ['crossref', 'jstor', 'degruyter', 'doaj', 'gbi']

        for source in sources:
            with self.input().get(source).open() as handle:
                for row in handle:
                    issns[source].add(row.strip())

        with self.output().open('w') as output:
            for issn in sorted(issns['file']):
                fields = []
                for source in sources:
                    if issn in issns[source]:
                        fields.append(source)
                if len(fields) == 0:
                    output.write_tsv(issn, "NOT_FOUND")
                else:
                    output.write_tsv(issn, "|".join(fields))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIISSNCoverageCatalogMatches(AITask):
    """
    For all ISSN, that are NOT_FOUND in AI, do a HTTP request to the current
    catalog and scrape the number of results.

    Example output:

    0171-4880   FOUND_RESULTS_1 https://katalog.ub.uni-...
    0171-5801   ERR_NOT_IN_CATALOG  https://katalog.ub....
    0171-5860   FOUND_RESULTS_466   https://katalog.ub....
    0171-7227   FOUND_RESULTS_2 https://katalog.ub.uni-...

    """
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def requires(self):
        return AICoverageISSN(date=self.date, isil=self.isil)

    def run(self):
        if not self.isil == 'DE-15':
            raise RuntimeError('not implemented except for DE-15')

        cache = URLCache(directory=os.path.join(tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=3)
        cache.sess.mount('http://', adapter)

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for i, row in enumerate(handle.iter_tsv(cols=('issn', 'status'))):
                    if row.status == 'NOT_FOUND':
                        link = 'https://katalog.ub.uni-leipzig.de/Search/Results?lookfor=%s&type=ISN' % row.issn
                        self.logger.info('fetch #%05d: %s' % (i, link))
                        body = cache.get(link)
                        if 'Keine Ergebnisse!' in body:
                            output.write_tsv(row.issn, 'ERR_NOT_IN_CATALOG', link)
                        else:
                            soup = BeautifulSoup(body)
                            rs = soup.findAll("div", {"class" : "floatleft"})
                            if len(rs) == 0:
                                output.write_tsv(row.issn, 'ERR_LAYOUT', link)
                                continue
                            first = rs[0]
                            match = re.search(r'Treffer([0-9]+)-([0-9]+)von([0-9]+)', first.text)
                            if match:
                                total = match.group(3)
                                output.write_tsv(row.issn, 'FOUND_RESULTS_%s' % total, link)
                            else:
                                output.write_tsv(row.issn, 'ERR_NO_MATCH', link)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIISSNCoverageSolrMatches(AITask):
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def requires(self):
        return AICoverageISSN(date=self.date, isil=self.isil)

    def run(self):
        if not self.isil == 'DE-15':
            raise RuntimeError('not implemented except for DE-15')

        cache = URLCache(directory=os.path.join(tempfile.gettempdir(), '.urlcache'))
        adapter = requests.adapters.HTTPAdapter(max_retries=3)
        cache.sess.mount('http://', adapter)

        finc = config.get('ai', 'finc-solr')
        ai = config.get('ai', 'ai-solr')

        with self.input().open() as handle:
            with self.output().open('w') as output:
                for i, row in enumerate(handle.iter_tsv(cols=('issn', 'status'))):
                    if row.status == 'NOT_FOUND':
                        link = '%s/select?q=institution:%s+AND+issn:%s&wt=json' % (finc, self.isil, row.issn)
                        self.logger.info('fetch #%05d: %s' % (i, link))
                        body = cache.get(link)
                        content = json.loads(body)
                        output.write_tsv('finc', row.issn, content['response']['numFound'], link)
                    else:
                        link = '%s/select?q=institution:%s+AND+issn:%s&wt=json' % (ai, self.isil, row.issn)
                        self.logger.info('fetch #%05d: %s' % (i, link))
                        body = cache.get(link)
                        content = json.loads(body)
                        output.write_tsv('ai', row.issn, content['response']['numFound'], link)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class AIISSNDetailedCoverageReport(AITask):
    """
    Detailed coverage report with Pandas (local dependency).
    """
    date = ClosestDateParameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def requires(self):
        return AIISSNCoverageSolrMatches(date=self.date, isil=self.isil)

    def run(self):
        import pandas as pd

        metrics = collections.defaultdict(dict)
        df = pd.read_csv(self.input().path, sep='\t', names=['shard', 'issn', 'count'], usecols=[0, 1, 2])

        # number of articles per shard
        metrics['sum']['finc'] = sum(df[df.shard == 'finc']['count'])
        metrics['sum']['ai'] = sum(df[df.shard == 'ai']['count'])

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class AIISSNCoverageReport(AITask):
    """
    For all ISILs, see what percentage of ISSN are covered in AI.

    Example report:

    DE-105  52.65
    DE-14   53.06
    DE-15   53.41
    DE-1972 43.99
    DE-8    52.94
    DE-Bn3  50.53
    DE-Brt1 43.99
    DE-Ch1  52.35
    DE-D117 43.99
    DE-D161 51.88
    DE-Gla1 51.88
    DE-Ki95 54.08
    DE-L229 50.83
    DE-Pl11 43.99
    DE-Rs1  51.11
    DE-Zi4  52.45

    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        prerequisite = AMSLHoldingsISILList(date=self.date)
        luigi.build([prerequisite])
        isils = set()
        with prerequisite.output().open() as handle:
            for isil in (string.strip(line) for line in handle):
                isils.add(isil)

        return dict((isil, AICoverageISSN(date=self.date, isil=isil)) for isil in isils)

    def run(self):
        with self.output().open('w') as output:
            for isil, target in sorted(self.input().iteritems()):
                counter = collections.Counter()
                with target.open() as handle:
                    for line in handle:
                        if 'NOT_FOUND' in line:
                            counter['miss'] += 1
                        else:
                            counter['hits'] += 1
                total = counter['hits'] + counter['miss']
                if total == 0:
                    continue
                ratio = (100.0 / total) * counter['hits']
                output.write_tsv(isil, '%0.2f' % ratio)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
[holdings]

sparql-endpoint = https://example.com/sparql-vt
"""

from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
import datetime
import json
import luigi
import re
import requests
import urllib

config = Config.instance()

def query_for_isil(isil):
    """
    For a given ISIL, return the path to the corresponding
    holdings file in AMSL.
    """
    if not re.match('[0-9a-zA-Z-]+', isil):
        raise RuntimeError('invalid ISIL: %s' % isil)

    return """
    prefix lobid: <http://purl.org/lobid/lv#>
    prefix amsl: <http://vocab.ub.uni-leipzig.de/amsl/>
    select ?path
    from <http://amsl.technology/consortial/>
    from <http://lobid.org/>
    where {
        ?bib amsl:linkToHoldingsFile ?path .
        ?bib lobid:isil ?isil .
        FILTER(STRENDS(?isil, "%s"))
    }
    """ % (isil)

class HoldingsTask(DefaultTask):
    TAG = 'holdings'

class HoldingsFile(HoldingsTask):
    """ Make a Holdings-File available for a given ISIL. """

    date = luigi.Parameter(default=datetime.date.today())
    isil = luigi.Parameter(default='DE-15')

    def run(self):
        encoded = urllib.urlencode({'default-graph-uri': '',
            'format': 'application/json', 'query': query_for_isil(self.isil)})

        url = "%s?%s" % (config.get('holdings', 'sparql-endpoint'), encoded)
        r = requests.get(url)
        if r.status_code >= 400:
            raise RuntimeError('%s on %s' % (r.status_code, url))

        response = json.loads(r.text)
        bindings = response['results']['bindings']

        if len(bindings) == 0:
            raise RuntimeError('no holdings found for %s' % self.isil)
        if len(bindings) > 1:
            raise RuntimeError('ambiguous holdings found for %s' % self.isil)

        holdings_url = bindings[0]['path']['value']
        output = shellout("""curl --fail "{url}" > {output}""", url=holdings_url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))

class HoldingsFinc(HoldingsTask, luigi.WrapperTask):
    """ Download finc-related holdings. """

    date = luigi.Parameter(default=datetime.date.today())

    def requires(self):
        isils = ['DE-15', 'DE-14', 'DE-105', 'DE-Ch1', 'DE-Gla1', 'DE-Bn3']
        return [HoldingsFile(isil=isil, date=self.date) for isil in isils]

    def output(self):
        return self.input()



# coding: utf-8
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from gluish.common import Executable
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import luigi

class HistbestTask(DefaultTask):
    TAG = '036'

    def closest(self):
        return monthly(date=self.date)

class HistbestHarvest(HistbestTask):
    """ OAI Harvest. """

    date = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default='oai_dc')
    url = luigi.Parameter(default='http://histbest.ub.uni-leipzig.de/oai2', significant=False)

    def requires(self):
        return Executable(name='oaimi', message='http://github.com/miku/oaimi')

    def run(self):
        output = shellout('oaimi -verbose -root records -until {until} -prefix {prefix} {url} > {output}',
                          prefix=self.prefix, url=self.url, until=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))



# coding: utf-8

"""
Helper for databases.
"""

from pymysql.cursors import SSCursor
import logging
import pymysql
import sqlite3
import urlparse

class sqlitedb(object):
    """
    Simple context manager for sqlite3 databases. Commits everything at exit.

        with sqlite3db('/tmp/test.db') as cursor:
            query = cursor.execute('SELECT * FROM items')
            result = query.fetchall()
    """
    def __init__(self, path, timeout=5.0, detect_types=0):
        self.path = path
        self.conn = None
        self.cursor = None
        self.timeout = timeout
        self.detect_types = detect_types

    def __enter__(self):
        self.conn = sqlite3.connect(self.path, timeout=self.timeout, detect_types=self.detect_types)
        self.conn.text_factory = str
        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_class, exc, traceback):
        self.conn.commit()
        self.conn.close()

class mysqldb(object):
    """ Context manager for MySQL database access.

        with mysqldb('mysql://user:pass@host/db', stream=True) as cursor:
            query = cursor.execute('SELECT * FROM items')
            result = query.fetchall()

    """
    def __init__(self, url, stream=False, commit_on_exit=False):
        result = urlparse.urlparse(url, scheme='mysql')
        self.hostname = result.hostname
        self.username = result.username
        self.password = result.password
        self.database = result.path.strip('/')
        self.stream = stream
        self.commit_on_exit = commit_on_exit
        self.conn = None
        self.cursor = None

    def __enter__(self):
        if self.stream:
            self.conn = pymysql.connect(host=self.hostname, user=self.username,
                                        passwd=self.password, db=self.database,
                                        cursorclass=SSCursor)
        else:
            self.conn = pymysql.connect(host=self.hostname, user=self.username,
                                        passwd=self.password, db=self.database)

        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_class, exc, traceback):
        if self.commit_on_exit:
            self.conn.commit()
        self.cursor.close()
        self.conn.close()


# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
IMDB.
"""

from siskin.common import FTPMirror
from gluish.intervals import hourly
from siskin.task import DefaultTask
import luigi

class IMDBTask(DefaultTask):
    TAG = 'imdb'

class IMDBSync(IMDBTask):

    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return FTPMirror(host='ftp.fu-berlin.de',
                         username='anonymous',
                         password='',
                         base='/pub/misc/movies/database/')

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Fuzzy deduplication related tasks.
"""

from elasticsearch import helpers as eshelpers
from siskin.benchmark import timed
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.utils import ElasticsearchMixin
import datetime
import elasticsearch
import luigi
import marcx
import tempfile

class FuzzyTask(DefaultTask):
    """ Fuzzy deduplication related things. """
    TAG = 'fuzzy'

    def closest(self):
        return weekly(self.date)

class FuzzyEditionList(FuzzyTask, ElasticsearchMixin):
    """ For a single index, create a TSV of the form

        Index, Id, Edition Statement

    where Edition Statement (250.a) is either NULL, a single string of multiple
    strings, separeted by a pipe (|).

    Index parameters can be adjusted like (space separated values in a single string):

        $ taskdo FuzzyEditionList --index ebl
    """
    date = ClosestDateParameter(default=datetime.date.today())
    index = luigi.Parameter(description='name of the index')

    @timed
    def run(self):
        es = elasticsearch.Elasticsearch([dict(host=self.es_host, port=self.es_port)])
        hits = eshelpers.scan(es, {'query': {'match_all': {}},
            'fields': ["content.250.a"]}, index=self.index,
            doc_type='title', scroll='10m', size=10000)
        with self.output().open('w') as output:
            for hit in hits:
                fields = hit.get('fields')
                if fields:
                    edition = 'NULL'
                    entries = fields.get('content.250.a', [])
                    if entries:
                        edition = '|'.join(entries).encode('utf-8')
                    output.write_tsv(hit['_id'], hit['_index'], edition)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class FuzzyEditionListRange(FuzzyTask, ElasticsearchMixin):
    """
    For a list of indices, create a FuzzyEditionList and concatenate them.

        $ taskdo FuzzyEditionList --indices "bsz ebl nep"
    """
    date = ClosestDateParameter(default=datetime.date.today())
    indices = luigi.Parameter(default="bsz ebl nep")

    def requires(self):
        indices = self.indices.split()
        for index in indices:
            yield FuzzyEditionList(date=self.date, index=index,
                                   es_host=self.es_host, es_port=self.es_port)

    @timed
    def run(self):
        _, output = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}", input=target.path, output=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class FuzzyFieldList(FuzzyTask, ElasticsearchMixin):
    """ Run estab to gather a TSV of (id, index, type, values...) for  comparisons. """
    date = ClosestDateParameter(default=datetime.date.today())
    indices = luigi.Parameter(description='indices to extract values from')
    fields = luigi.Parameter(default='_id _index _type content.245.a content.245.b', description='values to extract')

    def requires(self):
        return Executable(name='estab', message='http://git.io/bLY7cQ')

    @timed
    def run(self):
        output = shellout("""estab -host {host} -port {port} -indices "{indices}" -f "{fields}" > {output} """,
                          indices=self.indices,fields=self.fields, host=self.es_host, port=self.es_port)
        luigi.File(output).move(self.output().path)

    def output(self):
            return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class FuzzyTitlePool(FuzzyTask, ElasticsearchMixin):
    """ Deduplicate titles. """

    date = ClosestDateParameter(default=datetime.date.today())
    source = luigi.Parameter(description='indices that are compared')
    target = luigi.Parameter(description='indices that are compared against')
    null = luigi.Parameter(default="NOT_AVAILABLE", significant=False)

    def requires(self):
        return {'file': FuzzyFieldList(date=self.date, indices=self.source, fields='_id _index _type content.245.a content.245.b'),
                'esmlt': Executable(name='esmlt', message='http://git.io/ckXUgA')}

    @timed
    def run(self):
        # find similar titles
        output = shellout("""esmlt -host {host} -port {port} -indices "{target}" -fields "content.245.a content.245.b"
                             -file "{file}" -columns "4,5" > {output} """, host=self.es_host, port=self.es_port,
                             file=self.input().get('file').path, target=self.target)
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with luigi.File(output, format=TSV).open() as handle:
            with luigi.File(stopover, format=TSV).open('w') as output:
                for row in handle.iter_tsv(cols=('idl', 'il', 'tl', 'til', 'sl', 'idr', 'ir', 'tr', 'score', 'tir', 'sr')):
                    ml = ' '.join([v for v in (row.til, row.sl) if v and not v == "NOT_AVAILABLE"])
                    mr = ' '.join([v for v in (row.tir, row.sr) if v and not v == "NOT_AVAILABLE"])
                    if not ml: ml = "NOT_AVAILABLE"
                    if not mr: mr = "NOT_AVAILABLE"
                    output.write_tsv(row.idl, row.il, row.tl, ml,
                                     row.idr, row.ir, row.tr, mr, row.score)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)

class FuzzyCandidates(FuzzyTask):
    date = ClosestDateParameter(default=datetime.date.today())
    source = luigi.Parameter(description='indices that are compared')
    target = luigi.Parameter(description='indices that are compared against')
    null = luigi.Parameter(default="NOT_AVAILABLE", significant=False)

    measure = luigi.Parameter(default='ngram')
    threshold = luigi.FloatParameter(default=0.75)

    def requires(self):
        return {'file': FuzzyTitlePool(source=self.source, target=self.target, null=self.null, date=self.date),
                'stardust': Executable(name='stardust', message='http://git.io/_ehcwQ')}

    @timed
    def run(self):
        output = shellout("""stardust -f "4,8" {measure} {input} > {output}""", measure=self.measure, input=self.input().get('file').path)
        output = shellout(r"""/bin/bash -c "LANG=C sort -t$'\t' -k10,10 -nr {input} > {output}" """, input=output)
        output = shellout("""awk -F'\\t' '$10 > {threshold} {{print $0}}' {input} > {output} """, threshold=self.threshold, input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)



# coding: utf-8

"""
Twarc harvested data.
"""

from gluish.format import TSV
from gluish.utils import shellout
from siskin.configuration import Config
from siskin.task import DefaultTask
from siskin.utils import iterfiles
import datetime
import json
import luigi
import os
import tempfile

config = Config.instance()

class TwarcTask(DefaultTask):
    TAG = 'twarc'

class TwarcSync(TwarcTask):
    """
    Sync data from harvester.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        target = os.path.join(self.taskdir(), 'mirror')
        if not os.path.exists(target):
            os.makedirs(target)
        shellout("rsync -avzP {remote} {target}", remote=config.get('twarc', 'remote'), target=target)
        with self.output().open('w') as output:
            for path in iterfiles(target):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class TwarcChecksums(TwarcTask):
    """
    Calculate the checksum for each twarc file.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return TwarcSync(date=self.date)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='twarc-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                shellout("md5sum {input} >> {output}", input=row.path, output=stopover)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'))

class TwarcURLList(TwarcTask):
    """
    For a archive, identified by its checksum, extract all URLs.
    """
    checksum = luigi.Parameter()

    def requires(self):
        return TwarcChecksums()

    def run(self):
        archive = None

        with self.input().open() as handle:
            for line in handle:
                checksum, path = line.strip().split()
                if checksum == self.checksum:
                    archive = path

        if archive is None:
            raise RuntimeError('no archive with this checksum: %s' % self.checksum)

        with self.output().open('w') as output:
            with open(archive) as handle:
                for row in handle:
                    try:
                        tweet = json.loads(row)
                        for obj in tweet.get('entities', {}).get('urls', []):
                            output.write_tsv(obj.get('url').encode('utf-8'), obj.get('expanded_url').encode('utf-8'))
                    except Exception as err:
                        self.logger.warn(err)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class TwarcURLListWrapper(TwarcTask, luigi.WrapperTask):

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        prerequisite = TwarcChecksums()
        luigi.build([prerequisite])
        with prerequisite.output().open() as handle:
            for line in handle:
                checksum, path = line.strip().split()
                if not path.endswith('json'):
                    continue
                yield TwarcURLList(checksum=checksum)

    def output(self):
        return self.input()

class TwarcURLs(TwarcTask):

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return TwarcURLListWrapper(date=self.date)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}", input=target.path, output=stopover)
        luigi.File(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
BVB via OAI.
"""

from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.task import DefaultTask
import datetime
import luigi
import tempfile

class BVBTask(DefaultTask):
    TAG = '012'

    def closest(self):
        return monthly(date=self.date)

class BVBDump(BVBTask):
    """ Dump SSOAR. """

    def requires(self):
        return Executable(name='oaimi', message='https://github.com/miku/oaimi')

    def run(self):
        output = shellout("""
            cat <(echo '<collection xmlns="http://www.openarchives.org/OAI/2.0/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">')
                <(oaimi -verbose -set OpenData -prefix marc21 http://bvbr.bib-bvb.de:8991/aleph-cgi/oai/oai_opendata.pl)
                <(echo '</collection>') > {output}""")
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

class BVBCombine(BVBTask):
    """ Combine the chunks for a date range into a single file.
    The filename will carry a name, that will only include year
    and month for begin and end. """

    def requires(self):
        return SSOARDump()

    @timed
    def run(self):
        output = shellout("""yaz-marcdump -f utf-8 -t utf-8 -i marcxml -o marc {input} > {output}""", input=self.input().path, ignoremap={5: 'TODO: fix this'})
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class BVBJson(BVBTask):
    """ Json version. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SSOARCombine(date=self.date)

    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.closest())
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

# class BVBIndex(BVBTask, CopyToIndex):
#     begin = luigi.DateParameter(default=datetime.date(2011, 1, 1))
#     date = ClosestDateParameter(default=datetime.date.today())

#     index = "bvb"
#     doc_type = "title"
#     purge_existing_index = True

#     mapping = {'title': {'date_detection': False,
#                           '_id': {'path': 'content.001'},
#                           '_all': {'enabled': True,
#                                    'term_vector': 'with_positions_offsets',
#                                    'store': True}}}

#     def update_id(self):
#         """ This id will be a unique identifier for this indexing task."""
#         return self.effective_task_id()

#     def requires(self):
#         return BVBJson(begin=self.begin, date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
OSO.

This group of task worked once, but OSO completely redesigned their site,
so there tasks will fail now.
"""

from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from luigi.contrib.esindex import CopyToIndex
from siskin.benchmark import timed
from siskin.database import sqlitedb
from siskin.task import DefaultTask
from siskin.utils import copyregions, iterfiles
import BeautifulSoup
import datetime
import luigi
import os
import tempfile
import urlparse

class OSOTask(DefaultTask):
    """ Oxford Scholarship Online. """
    TAG = '018'

    def closest(self):
        return monthly(self.date)

class OSODownloadPage(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="http://www.universitypressscholarship.com/page/1250/oso-marc-archive",
                          significant=False)

    @timed
    def run(self):
        output = shellout("wget -q --retry-connrefused -O {output} '{url}' ", url=self.url)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='html'))

class OSOMarcFiles(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return OSODownloadPage(date=self.date)

    @timed
    def run(self):
        with self.input().open() as handle:
            soup = BeautifulSoup.BeautifulSoup(handle.read())
        with self.output().open('w') as output:
            for anchor in soup.findAll('a'):
                href = anchor.get('href')
                if not href.endswith('mrc'):
                    continue
                output.write_tsv(href)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class OSOMarcDownload(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return OSOMarcFiles(date=self.date)

    def run(self):
        base = "http://www.universitypressscholarship.com/"

        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                dirname, basename = row.path.split('/')[-2:]
                slugged = dirname.replace('%20', '-').lower()
                url = urlparse.urljoin(base, row.path)
                dst = os.path.join(self.taskdir(), '{0}-{1}'.format(slugged, basename))
                if os.path.exists(dst):
                    continue
                output = shellout("""wget --retry-connrefused "{url}" -O {output} """, url=url)
                luigi.File(output).move(dst)

        with self.output().open('w') as output:
            for path in iterfiles(self.taskdir()):
                if not path.endswith('mrc'):
                    continue
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)

class OSOCombine(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return OSOMarcDownload(date=self.date)

    def run(self):
        _, combined = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                shellout('cat {input} >> {output}', input=row.path, output=combined)
        luigi.File(combined).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'), format=TSV)

class OSOSeekmap(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return OSOCombine(date=self.date)

    def run(self):
        output = shellout('marcmap -o {output} {input}', input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='db'))

class OSOSurface(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return OSOCombine(date=self.date)

    def run(self):
        output = shellout("marctotsv {input} 001 005 > {output}",
                          input=self.input().path)
        output = shellout("LANG=C sort {input} > {output}", input=output)
        output = shellout("LANG=C tac {input} | LANG=C uniq -w 13 > {output}", input=output)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)

class OSOSnapshot(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'surface': OSOSurface(date=self.date),
                'seekmap': OSOSeekmap(date=self.date),
                'file': OSOCombine(date=self.date)}

    def run(self):
        with self.input().get('surface').open() as handle:
            with self.output().open('w') as output:
                with self.input().get('file').open() as fh:
                    with sqlitedb(self.input().get('seekmap').path) as cursor:
                        regions = []
                        for row in handle.iter_tsv(cols=('id', 'date')):
                            cursor.execute("SELECT offset, length FROM seekmap where id = ?", (row.id,))
                            regions.append(cursor.fetchone())
                        copyregions(fh, output, regions)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))

class OSOJson(OSOTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return OSOSnapshot(date=self.date)

    @timed
    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          date=self.date, input=self.input().path)
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class OSOIndex(OSOTask, CopyToIndex):
    date = ClosestDateParameter(default=datetime.date.today())

    index = 'oso'
    doc_type = 'title'
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return OSOJson(date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from siskin.benchmark import timed
from luigi.contrib.esindex import CopyToIndex
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask
import datetime
import luigi
import pymarc

class DOABTask(DefaultTask):
    TAG = '026'

    def closest(self):
        return monthly(date=self.date)

class DOABDump(DOABTask):
    """
    Unfiltered dump.
    """
    def requires(self):
        return Executable(name='oaimi', message='https://github.com/miku/oaimi')

    def run(self):
        output = shellout("""
            cat <(echo '<collection xmlns="http://www.openarchives.org/OAI/2.0/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">')
                <(oaimi -verbose -prefix marcxml http://www.doabooks.org/oai)
                <(echo '</collection>') > {output}""")
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))

class DOABCombine(DOABTask):
    """ Combine the chunks for a date range into a single file.
    The filename will carry a name, that will only include year
    and month for begin and end. """

    begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default="marcxml", significant=False)
    url = luigi.Parameter(default="http://www.doabooks.org/oai", significant=False)
    collection = luigi.Parameter(default=None, significant=False)

    def requires(self):
        return DOABDump()

    @timed
    def run(self):
        tmp = shellout("""yaz-marcdump -l 9=97 -f UTF8 -t UTF8
                          -i marcxml -o marc {input} > {output}""",
                          input=self.input().path, ignoremap={5: 'INVESTIGATE'})

        # filter dups, bad ids and unreadable records ...
        # TODO: make this possible with gomarckit as well
        seen = set()
        with open(tmp) as handle:
            with self.output().open('w') as output:
                reader = pymarc.MARCReader(handle, to_unicode=True)
                writer = pymarc.MARCWriter(output)
                while True:
                    try:
                        record = reader.next()
                        record_id = record['001'].value()
                        if record_id in ('15270', '15298', '15318', '15335'):
                            self.logger.debug("Skipping {0}".format(record_id))
                            continue
                        if not record_id in seen:
                            writer.write(record)
                            seen.add(record_id)
                        else:
                            self.logger.debug("Skipping duplicate: {0}".format(record_id))
                    except pymarc.exceptions.RecordDirectoryInvalid as err:
                        self.logger.warn(err)
                    except StopIteration:
                        break

    def output(self):
        """ Use monthly files. """
        return luigi.LocalTarget(path=self.path(digest=True, ext='mrc'))

class DOABJson(DOABTask):
    """ Convert to JSON. """

    begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return DOABCombine(begin=self.begin, date=self.date)

    def run(self):
        output = shellout("marctojson -m date={date} {input} > {output}",
                          input=self.input().path, date=self.date)
        luigi.File(output).move(self.output().path)

    def output(self):
        """ Use monthly files. """
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class DOABIndex(DOABTask, CopyToIndex):
    begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
    date = ClosestDateParameter(default=datetime.date.today())

    index = "doab"
    doc_type = "title"
    purge_existing_index = True

    mapping = {'title': {'date_detection': False,
                          '_id': {'path': 'content.001'},
                          '_all': {'enabled': True,
                                   'term_vector': 'with_positions_offsets',
                                   'store': True}}}

    def update_id(self):
        """ This id will be a unique identifier for this indexing task."""
        return self.effective_task_id()

    def requires(self):
        return DOABJson(begin=self.begin, date=self.date)



# coding: utf-8
# pylint: disable=C0103,W0232
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
Define a siskin wide task with artefacts under core.home directory.
"""

from gluish.task import BaseTask
from gluish.utils import shellout
from siskin.configuration import Config
import logging
import os
import random
import siskin

config = Config.instance()

class DefaultTask(BaseTask):
    """ A base task that sets its base directory based on config value. """
    BASE = config.get('core', 'home', Config.NO_DEFAULT)

    def assets(self, path):
        """ Return the absolute path to the asset. `path` is the relative path
        below the assets root dir. """
        return os.path.join(os.path.dirname(__file__), 'assets', path)

    @property
    def logger(self):
        # logging uses singleton internally, so no worries
        return logging.getLogger('siskin')

    @property
    def logfile(self):
        """ A logfile per task for stderr redirects. """
        name = self.taskdir().split('/')[-1]
        return os.path.join(config.get('core', 'logdir', '/var/log/siskin'), '%s.log' % name)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.task import DefaultTask
import BeautifulSoup
import collections
import datetime
import hashlib
import json
import luigi
import marcx
import pymarc
import tempfile

class QucosaTask(DefaultTask):
    TAG = '022'

    def closest(self):
        return monthly(date=self.date)

class QucosaDump(QucosaTask):

    def run(self):
        output = shellout("oaimi -verbose -prefix xMetaDissPlus > {output}")
        luigi.File(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())

# class QucosaHarvestChunk(OAIHarvestChunk, QucosaTask):
#     """ Harvest all files in chunks. """

#     begin = luigi.DateParameter(default=datetime.date.today())
#     end = luigi.DateParameter(default=datetime.date.today())
#     url = luigi.Parameter(default="http://www.qucosa.de/oai", significant=False)
#     prefix = luigi.Parameter(default="xMetaDissPlus")
#     collection = luigi.Parameter(default=None)

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='xml', digest=True))

# class QucosaJsonChunk(QucosaTask):
#     """ Convert Qucosa XML into JSON by hand. Any better way?
#     Maybe sth. like https://github.com/gawel/pyquery """
#     begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
#     end = luigi.DateParameter(default=datetime.date.today())

#     def requires(self):
#         return QucosaHarvestChunk(begin=self.begin, end=self.end)

#     @timed
#     def run(self):
#         docs = []

#         with self.input().open() as handle:
#             soup = BeautifulSoup.BeautifulStoneSoup(handle.read())
#             for record in soup.findAll('record'):
#                 doc = collections.defaultdict(list)

#                 # header
#                 header = record.find('header')
#                 if header.get('status') == 'deleted':
#                     continue

#                 # IDs
#                 doc['oai'] = header.find('identifier').text
#                 doc['id'] = doc['oai'].replace('oai:qucosa.de:', '')

#                 # the whole metadata element
#                 metadata = record.find('metadata')
#                 diss = metadata.find('xmetadiss:xmetadiss')
#                 if not diss:
#                     raise ValueError('qucosa without xmetadiss')

#                 # title
#                 if diss.find('dc:title'):
#                     doc['title'] = diss.find('dc:title').text
#                 else:
#                     self.logger.error("Qucosa w/o title: {0}".format(doc.get('id', 'NA')))
#                     # raise ValueError('qucosa without title')

#                 # subtitle
#                 alternative = diss.find('dcterms:alternative')
#                 if alternative and (alternative.get("xsi:type") ==
#                                     'ddb:talternativeISO639-2'):
#                     doc['alt'] = alternative.text

#                 # source
#                 source = diss.find('dc:source')
#                 if source and source.get('xsi:type') == 'ddb:noScheme':
#                     doc['source'] = source.text

#                 # creators
#                 for creator in diss.findAll('dc:creator'):
#                     forename = creator.find('pc:forename', recursive=True)
#                     surname = creator.find('pc:surname', recursive=True)
#                     doc['creator'].append("%s %s" % (forename.text, surname.text))

#                 # subjects
#                 for subject in diss.findAll('dc:subject'):
#                     doc[subject['xsi:type']].append(subject.text)

#                 # abstracts
#                 for abstract in diss.findAll('dc:abstract'):
#                     doc['abstract'].append(abstract.text)

#                 # tocs
#                 for toc in diss.findAll('dcterms:tableOfContents'):
#                     doc['toc'].append(toc.text)

#                 # publisher
#                 for publisher in diss.findAll('dc:publisher'):
#                     name = publisher.find('cc:name', recursive=True)
#                     place = publisher.find('cc:place', recursive=True)
#                     doc['publisher'].append("%s %s" % (name.text, place.text))

#                 # contributor
#                 for contributor in diss.findAll('dc:contributor'):
#                     forename = contributor.find('pc:forename', recursive=True)
#                     surname = contributor.find('pc:surname', recursive=True)
#                     role = contributor.get('thesis:role')
#                     doc['contributor'].append({
#                         'name': '%s %s' % (forename.text, surname.text),
#                         'role': role
#                     })

#                 # publication type
#                 if diss.find('dc:type').get('xsi:type') == 'dini:PublType':
#                     doc['pubtype'] = diss.find('dc:type').text

#                 # date issued
#                 try:
#                     doc['issued'] = diss.find('dcterms:issued').text
#                 except AttributeError:
#                     pass

#                 # language
#                 try:
#                     doc['lang'] = diss.find('dc:language').text
#                 except AttributeError:
#                     pass

#                 # urns
#                 for dcid in diss.findAll('dc:identifier'):
#                     if dcid.get('xsi:type') == 'urn:nbn':
#                         doc['urn'] = dcid.text
#                         doc['id'] = dcid.text # override id

#                 # transfer (broken urls?)
#                 for transfer in diss.findAll('ddb:transfer'):
#                     if transfer.get('ddb:type') == 'dcterms:URI':
#                         doc['transfer'] = transfer.text

#                 # medium
#                 for medium in diss.findAll('dcterms:medium'):
#                     if medium.get('xsi:type') == 'dcterms:IMT':
#                         doc['mimetype'] = medium.text

#                 doc['filenumber'] = int(diss.find('ddb:filenumber').text)

#                 doc['filesize'] = 0 # adding up
#                 for fileprop in diss.findAll('ddb:fileproperties'):
#                     try:
#                         doc['filesize'] += int(fileprop.get('ddb:filesize', 0))
#                         doc['filename'].append(fileprop.get('ddb:filename'))
#                     except ValueError:
#                         self.logger.warn("Empty/wrong filesize in {0}".format(doc['id']))

#                 for ispartof in diss.findAll('dcterms:ispartof'):
#                     if ispartof.get('xsi:type') == 'ddb:noScheme':
#                         doc['ispartof_title'].append(ispartof.text)
#                     if ispartof.get('xsi:type') == 'ddb:ISSN':
#                         doc['ispartof_issn'].append(ispartof.text)
#                     if ispartof.get('xsi:type') == 'ddb:URI':
#                         doc['ispartof_uri'].append(ispartof.text.replace(
#                                                    'http://nbn-resolving.de/', ''))

#                 docs.append(doc)

#         # dump as json, one document per line
#         with self.output().open('w') as output:
#             for doc in docs:
#                 output.write(json.dumps(doc))
#                 output.write('\n')

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='json'))

# class QucosaMarcChunk(QucosaTask):
#     """ Convert a jsonified qucosa XML into MARC.
#     TODO: use external standalone python script as asset. """

#     begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
#     end = ClosestDateParameter(default=datetime.date.today())

#     def requires(self):
#         return QucosaJsonChunk(begin=self.begin, end=self.end)

#     @timed
#     def run(self):
#         records = []
#         with self.input().open() as handle:
#             for line in handle:
#                 record = marcx.FatRecord(to_unicode=True, force_utf8=True)
#                 doc = json.loads(line)

#                 record.add('001', doc['id'])

#                 if 'lang' in doc:
#                     record.add('041', a=doc['lang'])

#                 if 'issued' in doc:
#                     deof = datetime.datetime.strptime(doc['issued'], '%Y-%m-%d').strftime('%y%m%d')
#                     year = datetime.datetime.strptime(doc['issued'], '%Y-%m-%d').strftime('%Y')
#                     record.add('008',
#                                data='{deof}s{year}    xx{skip}{lang}'.format(
#                                deof=deof, year=year, skip=18 * ' ',
#                                lang=doc.get('lang', '   ')))

#                 for i, name in enumerate(doc.get('creator', [])):
#                     if i == 0:
#                         record.add('100', a=name, e='author')
#                         continue
#                     record.add('700', a=name, e='author')

#                 for contributor in doc.get('contributor', []):
#                     record.add('700', a=contributor.get('name'),
#                                e=contributor.get('role'))

#                 if 'title' in doc:
#                     if 'alt' in doc:
#                         record.add('245', a=doc['title'], b=doc['alt'])
#                     else:
#                         record.add('245', a=doc['title'])

#                 for publisher in doc.get('publisher', []):
#                     record.add('260', b=publisher)

#                 if 'issued' in doc:
#                     year = datetime.datetime.strptime(doc['issued'], '%Y-%m-%d').strftime('%Y')
#                     record.add('260', c=year, _9=doc['issued'])

#                 if 'source' in doc:
#                     record.add('490', a=doc['source'])

#                 for toc in doc.get('toc', []):
#                     hashed = hashlib.sha1(toc.encode('utf-8')).hexdigest()
#                     record.add('520', indicators='2 ', _9=hashed)

#                 for abstract in doc.get('abstract', []):
#                     hashed = hashlib.sha1(abstract.encode('utf-8')).hexdigest()
#                     record.add('520', indicators='3 ', _9=hashed)

#                 if 'pubtype' in doc:
#                     record.add('500', a=doc['pubtype'])

#                 if 'url' in doc:
#                     record.add('856', indicators='40', u=doc['url'])

#                 if 'transfer' in doc:
#                     subs = collections.defaultdict(list)
#                     if 'mimetype' in doc:
#                         subs['q'] = doc.get('mimetype')
#                     if 'filesize' in doc and doc.get('filesize') > 0:
#                         subs['s'] = str(doc.get('filesize'))
#                     for fn in doc.get('filename', [])[:9]:
#                         if doc.get('fn'):
#                             subs['f'].append(doc.get(fn))

#                     if doc.get('filenumber') > 1:
#                         filenames = sorted(doc.get('filename'), reverse=True,
#                                            key=lambda fn: fn.endswith('.pdf'))
#                         fnstr = ', '.join(filenames[:3])
#                         if len(filenames) > 0:
#                             fnstr += ', ...'
#                         subs['_3'] = '%s (%s files: %s)' % (doc['transfer'],
#                                      doc['filenumber'], fnstr)

#                     subs['u'] = doc['transfer']
#                     record.add('856', indicators='41', **subs)

#                 for subject in doc.get('xMetaDiss:noScheme', []):
#                     if subject is not None:
#                         record.add('650', indicators=' 4', a=subject)

#                 for subject in doc.get('dcterms:DDC', []):
#                     record.add('082', indicators='0 ', a=subject)

#                 for subject in doc.get('xMetaDiss:RVK', []):
#                     record.add('084', indicators='  ', _2='rvk', a=subject)

#                 if 'ispartof_uri' in doc:
#                     subs = {}
#                     if 'ispartof_title' in doc:
#                         parts = doc.get('ispartof_title')[0].split(';')
#                         if len(parts) > 1:
#                             t = ';'.join(parts[:-1]).strip()
#                             v = parts[-1]
#                             if t and v:
#                                 subs['t'] = t
#                                 subs['v'] = v
#                         else:
#                             subs['a'] = doc.get('ispartof_title')[0]
#                     if 'ispartof_issn':
#                         subs['x'] = doc.get('ispartof_issn')[0]
#                     subs['_6'] = doc.get('ispartof_uri')[0]
#                     record.add('830', indicators=' 0', **subs)

#                 records.append(record)

#         with self.output().open('w') as output:
#             writer = pymarc.MARCWriter(output)
#             for record in records:
#                 writer.write(record)

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='mrc'))

# class QucosaCombine(QucosaTask):
#     """ Harvest qucosa, cut all dates to the first of month. Example:
#     2013-05-12 .. 2013-09-11 would be translated to
#     2013-05-01 .. 2013-09-01"""

#     begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
#     date = ClosestDateParameter(default=datetime.date.today())

#     def requires(self):
#         """ Only require up to the last full month. """
#         date = self.closest()
#         if date < self.begin:
#             raise RuntimeError('invalid range: %s - %s' % (self.begin, self.date))
#         dates = date_range(self.begin, date, 1, 'months')
#         for i, _ in enumerate(dates[:-1]):
#             yield QucosaMarcChunk(begin=dates[i], end=dates[i + 1])

#     @timed
#     def run(self):
#         _, combined = tempfile.mkstemp(prefix='tasktree-')
#         for target in self.input():
#             shellout("cat {input} >> {output}", input=target.path,
#                      output=combined)
#         output = shellout("marcuniq -o {output} {input}", input=combined)
#         luigi.File(output).move(self.output().path)

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='mrc'))

# class QucosaJson(QucosaTask):
#     """ Convert to JSON. """

#     begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
#     date = ClosestDateParameter(default=datetime.date.today())

#     def requires(self):
#         return QucosaCombine(begin=self.begin, date=self.date)

#     def run(self):
#         output = shellout("marctojson -m date={date} {input} > {output}",
#                           input=self.input().path, date=self.date)
#         luigi.File(output).move(self.output().path)

#     def output(self):
#         return luigi.LocalTarget(path=self.path(ext='json'))

# class QucosaIndex(QucosaTask, CopyToIndex):
#     begin = luigi.DateParameter(default=datetime.date(2010, 1, 1))
#     date = ClosestDateParameter(default=datetime.date.today())

#     index = 'qucosa'
#     doc_type = 'title'
#     purge_existing_index = True

#     mapping = {'title': {'date_detection': False,
#                           '_id': {'path': 'content.001'},
#                           '_all': {'enabled': True,
#                                    'term_vector': 'with_positions_offsets',
#                                    'store': True}}}

#     def update_id(self):
#         """ This id will be a unique identifier for this indexing task."""
#         return self.effective_task_id()

#     def requires(self):
#         return QucosaJson(begin=self.begin, date=self.date)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
A place for one time tasks and quick write-ups.
"""

from elasticsearch import helpers as eshelpers
from gluish.format import TSV
from siskin.sources.crossref import CrossrefDOIAndISSNList
from siskin.task import DefaultTask
import csv
import datetime
import elasticsearch
import json
import luigi
import marcx
import re

class AdhocTask(DefaultTask):
    """ Tasks that do not fit elsewhere. """
    TAG = 'adhoc'

class DOIList(AdhocTask):
    """
    For all open indices, run a heuristic DOI query and dump a TSV of
    (index, id, DOI, year, title and subtitle) tuples.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    encoding = luigi.Parameter(default='utf-8', significant=False)
    na = luigi.Parameter(default='NOT_AVAILABLE', significant=False)
    scroll = luigi.Parameter(default='10m', significant=False)
    timeout = luigi.IntParameter(default=30, significant=False)

    def run(self):
        es = elasticsearch.Elasticsearch(timeout=self.timeout)
        hits = eshelpers.scan(es, {'query': {"regexp": {"_all": "10\\.[0-9]{4,}"}}}, scroll=self.scroll)
        with self.output().open('w') as output:
            for hit in hits:
                doc_id, index = hit.get('_id'), hit.get('_index')
                content = hit.get('_source').get('content')
                matches = re.findall(r"10\.[0-9]{4,}/[^ \t\"]{3,}", json.dumps(content))
                for candidate in set(matches):
                    doc = marcx.marcdoc(hit)
                    title = (doc.values('245.a') or [self.na])[0]
                    subtitle = (doc.values('245.b') or [self.na])[0]
                    pubyear = (doc.values('260.c') or [self.na])[0]
                    output.write_tsv(index,
                                     doc_id,
                                     candidate.encode(self.encoding),
                                     pubyear.encode(self.encoding),
                                     title.encode(self.encoding),
                                     subtitle.encode(self.encoding))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)
